/*
 * _85_deco_cola_univ.c
 *
 * Created: 27/09/2015 12:12:16
 *  Author: German
 *
 *  Decodificador para las luces de cola empleando ATtiny85:
 *  - destelleantes, programable perodo y tiempo destello
 *  - en tringulo
 *       incandescentes, programable luminosidad
 *       petrleo con efecto fuego programable en intensidad tambin
 *  Activadas por funcin mapeable
 *  Emplea tres salidas pwm; prueba de decodificacin sin detener el timer
 *  Se podra emplear una cuarta salida para hacer la dependencia del sentido
 *      o bien utilizar charlieplexing para tener control sobre seis LEDs
 *  Cuarta salida para luces internas
 *
 *  -------------------------------------------------------------------------
 *      timer0 se utilizar para pwm en PB0 y PB1, y adems para DCC, en
 *          modo fast pwm. Tambin para intervalos
 *      timer1 para PB4
 */


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#define F_CPU 8000000UL

// pines empleados:
#define led0    PB0     // OC0A
#define led1    PB1     // OC0B
#define led2    PB4     // OC1B
#define ilum    PB2     // luz interna
#define p_dcc   PB3     // seal DCC

#define SI 0            // para el prembulo, la primera vez de la interrupcin
#define NO 1
#define T_FL_MX 5       // pasos de 10 ms de duracin del flash al mximo

#define ilum_on     PORTB |= (1 << ilum)
#define ilum_off    PORTB &= ~(1 << ilum)

// Alojar bits en GPIOR0
typedef struct              // aadir bits si son necesarios
{
    uint8_t bit0:1; uint8_t bit1:1; uint8_t bit2:1; uint8_t bit3:1;
    uint8_t bit4:1; uint8_t bit5:1; uint8_t bit6:1; uint8_t bit7:1;
}flags;

#define pqt_ok      ((volatile flags*)_SFR_MEM_ADDR(GPIOR0))->bit0  // paquete correcto
#define preamb      ((volatile flags*)_SFR_MEM_ADDR(GPIOR0))->bit1  // indica prembulo en la generacin de DCC
#define flag_10ms   ((volatile flags*)_SFR_MEM_ADDR(GPIOR0))->bit2  // activado cada 10 ms
#define bit_dcc     ((volatile flags*)_SFR_MEM_ADDR(GPIOR0))->bit3  // activado cada 10 ms
#define modo_svc    ((volatile flags*)_SFR_MEM_ADDR(GPIOR0))->bit4

// control de las funciones, una en cada bit:
//      ctrl_fn1 ->  0   0   0  F0    F4 F3 F2 F1
//      ctrl_fn2 -> F12 F11 F10 F9    F8 F7 F6 F5
uint8_t ctrl_fn1, ctrl_fn2, fn_nro;     // ********* quizs mejor en main()
// array para el peso de la funcin en su variable:
const uint8_t fn_peso[] = {16, 1, 2, 4, 8, 1, 2, 4, 8, 16, 32, 64, 128};
uint8_t buf_dcc[6];                     // buffer del ltimo comando recibido
volatile uint8_t nro_rnd;               // nmero random

// eeprom
const uint8_t cvs[] EEMEM =
{
    2,      // cv1 direccin byte bajo
    0,      // cv2 direccin byte alto
    0,      // cv3 nro de funcin que activa las luces de cola (led0, led1 y led2 segn configuracin)
    1,      // cv4 nro de funcin que activa la luz interior (ilum)
    0,      // cv5 configuracin de las luces: 0->flash laterales 1->tres incandescentes 2->tres petrleo
    0,      // cv6 programas predefinidos (al final)
    1,13,   // cv7-8 versin y fabricante (DIY=13)
    128,    // cv9 brillo mximo de laterales (led0 y led1)
    128,    // cv10 brillo mximo de inferior (led2)
    100,    // cv11 perodo de flash x 10 ms (50 a 255)
    2,      // cv12 pasos de apagado flash cada 10 ms (1 a cv11)
    30,     // cv13 porcentaje efecto fuego faroles laterales
    30      // cv14 porcentaje efecto fuego farol inferior
};
/*  cv13 programas predefinidos:
        0-> ninguno (usa valores de cvs)
        1->flash T=2 seg (200), brillo mx (255), apagado lento (2)
        2->flash T=1'5 seg (150), brillo mx (255), apagado rpido (15)
        3->tres incandescentes brillo medio (80)
        4->tres petrleo brillo bajo (50), efecto fuego 40%
        5->tres petrleo, laterales brillo muy bajo (25), central brillo medio (50) efecto fuego 80%
*/
// valores de fbrica, para reset
const uint8_t cvs_fab[] = {2,0,0,1,0,0,1,13,128,128,100,2,30,30};
// para los subndices:
enum {DIRL, DIRH, F_FAROL, F_ILUM, CONFIG, PRG_PDF, VER, FAB, BR_MAX_L, BR_MAX_I, T_FLASH, STEP_DWN, EF_FGO_L, EF_FGO_I};
// descartada la posibilidad de programar direccin larga con cv17-cv18 para multimaus

// interrupcin externa por cambio de nivel para decodificar la seal DCC
// Cuenta el tiempo del nivel alto con el timer0 como la diferencia de lecturas
//    entre el flanco de subida y el de bajada, sin detenerlo, para utilizarlo
//    en la generacin del pwm en PB0 y PB1
ISR(PCINT0_vect)
{
static uint8_t tmp_pulso;                           // valor inicial y final del timer0
static uint8_t aux_dcc, x_or, idx, num_bits;
    if (PINB & (1<<p_dcc)) tmp_pulso = TCNT0;       // flanco de subida inicio de cuenta
    else                                            // cuando pasa a nivel bajo
    {
        tmp_pulso = TCNT0 - tmp_pulso;              // flanco de bajada final de cuenta, duracin del pulso
        if (tmp_pulso>90) bit_dcc = 0;              // sin buscar exactitud: duracin mayor de 90 us => cero
        else bit_dcc = 1;                           //    y menor uno
        if (preamb == SI)                           // prembulo
        {
            if (bit_dcc)                            // otro '1'
            {
                if (num_bits) num_bits--;
            }
            else
            {
                if (!num_bits)                      // el '0' de inicio de datos
                {
                    preamb = NO;
                    num_bits = 9;
                    x_or = 0;
                    idx = 0;
                }
                else                                // no se cumple la trama, vuelta a inicio
                    num_bits = 10;
            }
        }
        else                // recepcin de los bytes
        {
            if (--num_bits) aux_dcc = aux_dcc * 2 + bit_dcc;
            else                                    // el separador de bytes
            {
                if (!bit_dcc)                       // cero, a por el siguiente byte
                {
                    buf_dcc [idx++] = aux_dcc;
                    x_or ^= aux_dcc;                // actualiza la x_or
                    num_bits = 9;
                }
                else                                // uno, fin del paquete
                {
                    preamb = SI;
                    num_bits = 10;                  // vuelta a empezar
                    if (x_or == aux_dcc) pqt_ok = 1;    // paquete correcto

                }
            }
        }
    }
    nro_rnd++;          // para tener un nmero aleatorio
}

// prescaler 1/8, luego cada 256 us
ISR(TIM0_OVF_vect, ISR_NOBLOCK)
{
static uint8_t cont=39;
    if(!--cont)             // 9984 us => 10 ms
    {
        flag_10ms = 1;
        cont = 39;
    }
}

// configuracin inicial para ATtiny85
void setup()
{
    DDRB = (1<<led0) | (1<<led1) | (1<<led2) | (1<<ilum);   // salidas
    GIMSK |= (1<<PCIE);                     // habilita interrupcin por cambio de nivel
    PCMSK |= (1<<PCINT3);                   // interrupcin por cambio de PB3
    // mode #3 fast pwm:
    TCCR0A = (2<<COM0A0) | (2<<COM0B0) | (3<<WGM00);
    OCR0A = 0;                              // inicialmente apagados
    OCR0B = 0;
    TCCR0B = (2 << CS00);                   // prescaler 1/8 (pasos de 1 us)
    TCCR1 = (4 << CS10) | (3 << COM1A0);    // prescaler Timer1 a 1/8
    GTCCR = (1 << PWM1B) | (2 << COM1B0);   // fast pwm en PB4
    OCR1B = 0;
    TIMSK |= (1 << TOIE0);          // habilita interrupcin por overflow (temporizacin)

    sei();                          // habilitador general
}

int main(void)
{
uint8_t cv[sizeof (cvs)];
uint8_t peso_fn_ilum, peso_fn_farol;
uint8_t ct_led0=0, ct_led1=0;           // contadores para los pwm de los leds
uint16_t direc, dir_larga;
uint8_t ef_fgo_l, ef_fgo_i, orden=0;  // para fuego

    setup();
    // mira si hay un 'reset de fbrica'
    orden = eeprom_read_byte((uint8_t*) FAB);             // pide reset si est a cero
    if (!orden)                                           // hay que resetear a valores originales
    {
        for (uint8_t i=0; i<sizeof(cv); i++)            // reinicia todas las cvs
        {
            while (!eeprom_is_ready());
            eeprom_update_byte((uint8_t *)((uint16_t) i), cvs_fab[i]);
        }
    }
    // Lee las cvs
    for (uint8_t i=0; i<sizeof(cv); i++) cv[i] = eeprom_read_byte((uint8_t*) (int) i);
    direc = cv[DIRL] + cv[DIRH]*100;	// direccin en decimal (se supone bien grabada)

    switch (cv[PRG_PDF])    // si es distinta de cero hay predefinido en estos casos
    {
        case 1: // 1->flash  lento
            cv[CONFIG] = 0;         // laterales flash
            cv[T_FLASH] = 200;      // 2 s
            cv[BR_MAX_L] = 255;     // brillo mximo
            cv[STEP_DWN] =2;        // apagado lento
            break;
        case 2: // 2->flash rpido
            cv[CONFIG] = 0;         // laterales flash
            cv[T_FLASH] = 150;      // 1'5 s
            cv[BR_MAX_L] = 255;     // brillo mximo
            cv[STEP_DWN] =10;       // apagado rpido
            break;
        case 3: // 3->tres incandescentes brillo medio
            cv[CONFIG] = 1;         // tres incandescentes
            cv[BR_MAX_L] = 60;
            cv[BR_MAX_I] = 60;
            break;
        case 4: // 4->tres petrleo brillo bajo
            cv[CONFIG] = 2;         // tres petrleo
            cv[BR_MAX_L] = 40;      // brillo bajo
            cv[BR_MAX_I] = 40;      
            cv[EF_FGO_L] = 40;      // fuego laterales 40%
            cv[EF_FGO_I] = 40;      // fuego inferior 40%
            break;
        case 5: // 5->tres petrleo, laterales brillo muy bajo, inferior brillo medio
            cv[CONFIG] = 2;         // tres petrleo
            cv[BR_MAX_L] = 20;      // brillo muy bajo
            cv[BR_MAX_I] = 40;      // brillo bajo
            cv[EF_FGO_L] = 40;      // fuego laterales 40%
            cv[EF_FGO_I] = 40;      // fuego inferior 40%
            break;
    }
    if (cv[T_FLASH] < 50) cv[T_FLASH] = 50;     // mnimo perodo
    if (cv[STEP_DWN] > cv[T_FLASH]) cv[STEP_DWN] = cv[T_FLASH]; // para restar correctamente
    if (cv[EF_FGO_L] > 100) cv[EF_FGO_L] = 100; // mxima profundidad del fuego
    if (cv[EF_FGO_I] > 100) cv[EF_FGO_I] = 100;
    // en los dems casos usa las cvs normales
    peso_fn_ilum = fn_peso[cv[F_ILUM]];         // peso de las funciones para buscarlas en ctrl_fni
    peso_fn_farol = fn_peso[cv[F_FAROL]];
    // ahora calcula el valor del efecto fuego a aplicar, para evitar clculos repetitivos luego en el bucle
    ef_fgo_i = (uint16_t)((cv[BR_MAX_I] * cv[EF_FGO_I]) / 100);     // mximo de profundidad del
    ef_fgo_l = (uint16_t)((cv[BR_MAX_L] * cv[EF_FGO_L]) / 100);     //      efecto fuego
    if (!ef_fgo_i) ef_fgo_i = 1;        // si fuera 0 que quede como incandescente
    if (!ef_fgo_l) ef_fgo_l = 1;
    while(1)
    {
        if (pqt_ok)         // decodificacin para ver las funciones activadas solamente
        {
            pqt_ok = 0;
            if (!buf_dcc[0]) modo_svc = 1;                                      // direccin 0 -> reset
            else
            {
                if ((buf_dcc[0] == cv[DIRL]) && !cv[DIRH])                      // direccin corta
                {
                    if ((buf_dcc[1] & 0xe0) == 0x80)                            // Function Group One:  100F0 F4F3F2F1
                        ctrl_fn1 = buf_dcc[1] & 0x1f;                           // funciones
                    else if ((buf_dcc[1] & 0xf0) == 0xa0)                       // funciones grupo 2: 101S DDDD (S=1->D8..5 // S=0->D12..9)
                        ctrl_fn2 = (ctrl_fn2 & 0x0f) | ((buf_dcc[1] & 0x0f) << 4); // D12..D9 en nibble alto
                    else if ((buf_dcc[1] & 0xf0) == 0xb0)                       // D8..D5
                        ctrl_fn2 = (ctrl_fn2 & 0xf0) | (buf_dcc[1] & 0x0f);     // pone D8..D5 en nibble bajo
                }
                // direccin de decodificador de accesorios o prog modo directo
                // en modo servicio: verificacin
                else if ((buf_dcc[0] == 0x74) && modo_svc)      // verificacin
                {
                    for (uint8_t i=0; i<sizeof(cv); i++)  // ltimo elemento del array
                    {
                        if ((buf_dcc[1] == i) && (buf_dcc[2] == cv[i]))    // si es alguna de las cvs utilizadas
                        {
                            DDRB = 0x17; OCR0A = 0xff; OCR0B = 0xff; OCR1B = 0xff; while(1);
                        }
                    }
                }
                // en modo servicio: programacin
                else if ((buf_dcc[0] == 0x7c) && modo_svc)      // escritura
                {   // graba si es alguna de las cvs que utiliza
                    for (uint8_t i=0; i<sizeof(cv); i++)
                    {
                        if (buf_dcc[1] == i)                        // si es alguna de las cvs utilizadas
                        {
                            if (buf_dcc[1]==VER) continue;          // no modifica versin
                            if (buf_dcc[1]==FAB)
                            {
                                if(buf_dcc[2]==8)                   //  indica hacer reset de cvs a los valores originales
                                {
                                    eeprom_update_byte((uint8_t *)((uint16_t) 7), 0);   // pone 0 en cv8 para detectarlo despus del setup()
                                    DDRB = 0x17; OCR0A = 0xff; OCR0B = 0xff; OCR1B = 0xff; while(1);
                                }
                                else continue;
                            }
                            eeprom_update_byte((uint8_t *)((uint16_t)buf_dcc[1]), buf_dcc[2]);
                            DDRB = 0x17; OCR0A = 0xff; OCR0B = 0xff; OCR1B = 0xff; while(1);
                        }
                    }
                }
                // si es direccin larga del decoder o de locomotora
                else if ((buf_dcc[0] > 191) && (buf_dcc[0] < 232))       		// entre 192 y 254: direccin larga
                {
                    dir_larga = (buf_dcc[0] & 0x3f)*256 | buf_dcc[1];
                    if (direc == dir_larga)                 // si es el decoder ver sus funciones
                    {
                        if ((buf_dcc[2] & 0xe0) == 0x80)   // Function Group One:  100F0 F4F3F2F1
                            ctrl_fn1 = buf_dcc[2] & 0x1f;                      // funciones
                        else if ((buf_dcc[2] & 0xf0) == 0xa0)  // funciones grupo 2: 101S DDDD (S=1->D8..5 // S=0->D12..9)
                            ctrl_fn2 = (ctrl_fn2 & 0x0f) | ((buf_dcc[2] & 0x0f) << 4); // D12..D9 en nibble alto
                        else if ((buf_dcc[2] & 0xf0) == 0xb0)                   // D8..D5
                            ctrl_fn2 = (ctrl_fn2 & 0xf0) | (buf_dcc[2] & 0x0f); // pone D8..D5 en nibble bajo
                    }
                }
                modo_svc = 0;                                               // modo servicio desactivado
            }
        }

        // ahora operar cada 10 ms segn las funciones activas
        if (flag_10ms && !modo_svc) // que no entre mientras est sobre 'resets' por si sigue programacin de cvs
        {
            flag_10ms = 0;
            if (cv[F_ILUM] < 5)            // est en ctrl_fn1
            {
                if (ctrl_fn1 & peso_fn_ilum) ilum_on;	// si est activo su bit (peso) enciende
                else ilum_off;
            }
            else if (cv[F_ILUM] < 13)      // est en ctrl_fn2
            {
                if (ctrl_fn2 & peso_fn_ilum) ilum_on;
                else ilum_off;
            }
            // condicin compleja para hacerlo todo aqu:
            // si la cv es <5 y est en ctrl_fn1; o bien es >4 y <13 y est en ctrl_fn2
            if (((cv[F_FAROL] < 5) && (ctrl_fn1 & peso_fn_farol)) || ((cv[F_FAROL] > 4) && (cv[F_FAROL] < 13) && (ctrl_fn2 & peso_fn_farol)))
            {   // enciende los faroles, dependiendo de su configuracin:
                if (!cv[CONFIG])        // laterales flash
                {
                    OCR1B = 0;          // inferior apagado
                    /*     T_FL_MX               // los dos laterales flash
                          |-------\         |-----
                          |        \        |
                        __|       | \_______|
                        ct_=0            T_FLASH
                    */
                    if (ct_led0 == cv[T_FLASH]) ct_led0 = 0;        // perodo
                    if (ct_led0 < T_FL_MX)
                    {
                        DDRB |= 1<<led0;            // salida
                        OCR0A = cv[BR_MAX_L];       // flash al mximo
                    }
                    else
                    {
                        if (OCR0A < STEP_DWN)
                        {
                            DDRB &= ~(1<<led0);     // entrada, para que se apague
                            OCR0A = 0;              // menor que los pasos, lo apaga
                        }
                        else OCR0A -= cv[STEP_DWN]; // desciende uno
                    }
                    if (ct_led1 == cv[T_FLASH]-4) ct_led1 = 0;      // perodo algo diferente del otro
                    if (ct_led1 < T_FL_MX)
                    {
                        DDRB |= 1<<led1;            // salida
                        OCR0B = cv[BR_MAX_L];       // duracin del pico de flash
                    }
                    else
                    {
                        if (OCR0B < STEP_DWN)
                        {
                            DDRB &= ~(1<<led1);     // entrada, para que se apague
                            OCR0B = 0;              // menor que los pasos, lo apaga
                        }
                        else OCR0B -= cv[STEP_DWN]; // desciende uno
                    }
                    ct_led0++;
                    ct_led1++;
                }
                else if (cv[CONFIG] == 1)  // incandescentes tringulo
                {
                    DDRB |= (1<<led0) | (1<<led1);  // salidas
                    OCR0A = cv[BR_MAX_L];
                    OCR0B = cv[BR_MAX_L];
                    OCR1B = cv[BR_MAX_I];
                }
                else if (cv[CONFIG] == 2)  // petrleo (efecto fuego) tringulo
                {
                    DDRB |= (1<<led0) | (1<<led1);  // salidas
                    if (!orden) OCR0A = cv[BR_MAX_L] - (nro_rnd % ef_fgo_l);
                    else if (orden == 1) OCR0B = cv[BR_MAX_L] - (nro_rnd % ef_fgo_l);
                    else OCR1B = cv[BR_MAX_I] - (nro_rnd % ef_fgo_i);
                    if (++orden > 2) orden = 0;
                }
            }
            else                      // estn apagados
            {
                DDRB &= ~(1<<led0) & ~(1<<led1);    // entradas para evitar el mnimo
                OCR0A = 0; OCR0B = 0; OCR1B = 0;
            }
        }
    }
}
